from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError, ImproperlyConfigured, ObjectDoesNotExist
from django import forms
from django.forms import ChoiceField, Field
from django.utils.translation import gettext as _


class GenericModelChoiceIterator:
    def __init__(self, field):
        self.field = field
        self.querysets = field.querysets

    def __iter__(self):
        if self.field.empty_label is not None:
            yield ("", self.field.empty_label)
        querysets = [qs.all() for qs in self.querysets]

        for queryset in querysets:
            # Can't use iterator() when queryset uses prefetch_related()
            if not queryset._prefetch_related_lookups:
                queryset = queryset.iterator()
            for obj in queryset:
                yield self.choice(obj)

    def __len__(self):
        return sum([len(qs) for qs in self.querysets]) + (1 if self.field.empty_label is not None else 0)

    def choice(self, obj):
        return self.field.prepare_value(obj), self.field.label_from_instance(obj)


class GenericModelChoiceField(forms.ModelChoiceField):
    iterator = GenericModelChoiceIterator

    def __init__(self, querysets, empty_label=_("Send to everyone"),
                 required=True, widget=None, label=None, initial=None,
                 help_text='', limit_choices_to=None,
                 **kwargs):
        if required and (initial is not None):
            self.empty_label = None
        else:
            self.empty_label = empty_label

        # Call Field instead of ChoiceField __init__() because we don't need
        # ChoiceField.__init__().
        Field.__init__(
            self, required=required, widget=widget, label=label,
            initial=initial, help_text=help_text, **kwargs
        )

        self._querysets = querysets

    def __deepcopy__(self, memo):
        result = super(ChoiceField, self).__deepcopy__(memo)
        # Need to force a new ModelChoiceIterator to be created, bug #11183
        if self.querysets is not None:
            result.querysets = [qs.all() for qs in self.querysets]
        return result

    def _get_querysets(self):
        if callable(self._querysets):
            querysets = self._querysets()
        else:
            querysets = self._querysets

        if len(querysets) > len(set([qs.model for qs in querysets])):
            raise ImproperlyConfigured()

        return querysets

    def _set_querysets(self, querysets):
        self._querysets = querysets
        self.widget.choices = self.choices

    querysets = property(_get_querysets, _set_querysets)

    def get_limit_choices_to(self):
        return None

    def get_queryset_for_content_type(self, content_type_id):
        content_type = ContentType.objects.get_for_id(content_type_id)

        for queryset in self.querysets:
            if queryset.model == content_type.model_class():
                return queryset

        raise ValidationError(self.error_messages['invalid_choice'], code='invalid_choice')

    def prepare_value(self, value):
        if hasattr(value, '_meta'):
            return '%s-%s' % (ContentType.objects.get_for_model(value).pk, value.pk)

        return value

    def to_python(self, value):
        if value in self.empty_values:
            return None
        try:
            ct_key = value.split('-')[0]
            object_key = value.split('-', 1)[1]
            queryset = self.get_queryset_for_content_type(ct_key)
            return queryset.get(pk=object_key)
        except (ValueError, TypeError, ObjectDoesNotExist):
            raise ValidationError(self.error_messages['invalid_choice'], code='invalid_choice')



from django.core import mail
from django.test import TestCase

from nuntius.models import Campaign, BaseSubscriber
from nuntius.tasks import send_campaign
from tests.models import TestSegment, TestSubscriber


class SendingTestCase(TestCase):
    fixtures = ['subscribers.json']

    def test_segments(self):
        self.assertEqual(TestSegment.objects.get(id="subscribed").get_subscribers_queryset().count(), 2)

    def test_send_campaign(self):
        segment = TestSegment.objects.get(id="subscribed")
        campaign = Campaign.objects.create(segment=segment)
        send_campaign(campaign.pk)

        self.assertEqual(segment.get_subscribers_queryset().count(), len(mail.outbox), campaign.get_sent_count())

    def test_send_campaign_without_segment(self):
        campaign = Campaign.objects.create()
        send_campaign(campaign.pk)

        self.assertEqual(TestSubscriber.objects.filter(subscriber_status=BaseSubscriber.STATUS_SUBSCRIBED).count(),
                         len(mail.outbox), campaign.get_sent_count())

    def test_send_only_to_subscribed(self):
        segment = TestSegment.objects.get(id="all_status")
        campaign = Campaign.objects.create(segment=segment)
        send_campaign(campaign.pk)

        self.assertEqual(
            segment.get_subscribers_queryset().filter(subscriber_status=BaseSubscriber.STATUS_SUBSCRIBED).count(),
            len(mail.outbox),
            campaign.get_sent_count()
        )

    def test_send_only_once(self):
        segment = TestSegment.objects.get(id="subscribed")
        campaign = Campaign.objects.create(segment=segment, message_content_text='test [email] test')
        send_campaign(campaign.pk)
        send_campaign(campaign.pk)

        self.assertEqual(segment.get_subscribers_queryset().count(), len(mail.outbox), campaign.get_sent_count())
        self.assertIn(f'test {segment.get_subscribers_queryset().first().email} test', str(mail.outbox[0].message()))






# Generated by Django 2.0.6 on 2018-06-30 04:54
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import nuntius.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.NUNTIUS_SUBSCRIBER_MODEL),
        ('contenttypes', '0002_remove_content_type_name')
    ]

    operations = [
        migrations.CreateModel(
            name='Campaign',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_uuid', models.UUIDField(blank=True, db_index=True, default=None, null=True, verbose_name='Celery tasks identifier')),
                ('name', models.CharField(max_length=255, verbose_name='Name (invisible to subscribers)')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('message_from_name', models.CharField(blank=True, max_length=255, verbose_name='"From" name')),
                ('message_from_email', models.EmailField(max_length=255, verbose_name='"From" email address')),
                ('message_reply_to_name', models.CharField(blank=True, max_length=255, verbose_name='"Reply to" name')),
                ('message_reply_to_email', models.CharField(blank=True, max_length=255, verbose_name='"Reply to" email address')),
                ('message_subject', models.CharField(blank=True, max_length=255, verbose_name='Message subject line')),
                ('message_content_html', models.TextField(blank=True, verbose_name='Message content (HTML)')),
                ('message_content_text', models.TextField(verbose_name='Message content (text)')),
                ('segment_id', models.CharField(max_length=255)),
                ('status', models.IntegerField(choices=[(0, 'Waiting'), (1, 'Sending'), (2, 'Sent'), (3, 'Error')], default=0)),
                ('segment_content_type', models.ForeignKey(limit_choices_to=nuntius.models.segment_cts_q, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='CampaignSentEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254, verbose_name='Email adress at event time')),
                ('datetime', models.DateTimeField(auto_now_add=True)),
                ('result', models.CharField(default='P', max_length=2, verbose_name='Operation result')),
                ('campaign', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='Campaign', to='nuntius.Campaign')),
                ('subscriber', models.ForeignKey(
                    null=True, blank=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name='Subscriber',
                    to=settings.NUNTIUS_SUBSCRIBER_MODEL
                )),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='campaignsentevent',
            unique_together={('campaign', 'subscriber')},
        ),
    ]



# Generated by Django 2.0.7 on 2018-07-18 08:28

from django.db import migrations, models
import django.db.models.deletion
import stdimage.models


class Migration(migrations.Migration):

    dependencies = [
        ('nuntius', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='MosaicoImage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('file', stdimage.models.StdImageField(upload_to='mosaico')),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='campaign',
            name='message_mosaico_data',
            field=models.TextField(blank=True, verbose_name='Mosaico data'),
        )
    ]



import os

from django.conf.global_settings import DEFAULT_FILE_STORAGE

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DEBUG = True
SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'tests',
    'nuntius'
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'nuntius',
    }
}

STATIC_URL = '/static/'

NUNTIUS_SEGMENT_MODELS = ['tests.testsegment']
NUNTIUS_SUBSCRIBER_MODEL = 'tests.TestSubscriber'
NUNTIUS_CELERY_BROKER_URL = 'redis://'

MEDIA_URL = '/media/'
MEDIA_ROOT = 'media'

EMAIL_HOST = 'localhost'
EMAIL_PORT = '1025'



from django.test import TestCase

from nuntius.models import BaseSubscriber
from tests.models import TestSubscriber


class SimpleSubscriberTestCase(TestCase):
    def test_use_email_property(self):
        subscriber = TestSubscriber(email='email@example.com')
        self.assertEqual(subscriber.get_subscriber_email(), 'email@example.com')

    def test_use_status_property(self):
        subscriber = TestSubscriber(email='email@example.com', subscriber_status=BaseSubscriber.STATUS_SUBSCRIBED)
        self.assertEqual(subscriber.get_subscriber_status(), BaseSubscriber.STATUS_SUBSCRIBED)
        subscriber.set_subscriber_status(BaseSubscriber.STATUS_UNSUBSCRIBED)
        self.assertEqual(subscriber.get_subscriber_status(), BaseSubscriber.STATUS_UNSUBSCRIBED)



from urllib.parse import urlparse

from PIL import Image
from django.conf import settings
from django.http import HttpResponse, HttpResponseBadRequest
from django.http.request import validate_host
from django.views.generic import DetailView

from nuntius.models import MosaicoImage
from nuntius.utils import generate_placeholder


class MosaicoImageProcessorView(DetailView):
    def get(self, *args, **kwargs):
        if self.request.GET.get('method') == 'placeholder':
            (width, height) = self.request.GET.get('params').split(',')
            image = generate_placeholder(int(width), int(height))
            response = HttpResponse(content_type="image/png")
            image.save(response, "PNG")
            return response

        if self.request.GET.get('src'):
            src = self.request.GET.get('src')
            host = urlparse(src).netloc.split(':')[0]
            allowed_hosts = settings.ALLOWED_HOSTS
            if settings.DEBUG and not allowed_hosts:
                allowed_hosts = ['localhost', '127.0.0.1', '[::1]']
            if not validate_host(host, allowed_hosts):
                return HttpResponseBadRequest()

            (width, height) = self.request.GET.get('params').split(',')
            (width, height) = (int(width.replace('null', '0')), int(height.replace('null', '0')))
            image = MosaicoImage.objects.get(file=urlparse(src).path.replace(settings.MEDIA_URL, '', 1))
            image = Image.open(image.file.path)

            if width and not height:
                ratio = width/image.size[0]
            if height and not width:
                ratio = height/image.size[1]
            if width and height:
                ratio = min(width/image.size[0], height/image.size[1])

            image.resize((round(size*ratio) for size in image.size), Image.ANTIALIAS)
            response = HttpResponse(content_type=f"image/{image.format.lower()}")
            image.save(response, image.format)
            return response

        return HttpResponseBadRequest()



# Generated by Django 2.0.7 on 2018-07-20 10:03

from django.db import migrations, models
import django.db.models.deletion
import nuntius.models


class Migration(migrations.Migration):

    dependencies = [
        ('nuntius', '0002_add_mosaico'),
    ]

    operations = [
        migrations.AlterField(
            model_name='campaign',
            name='segment_content_type',
            field=models.ForeignKey(limit_choices_to=nuntius.models.segment_cts_q, null=True, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType'),
        ),
        migrations.AlterField(
            model_name='campaign',
            name='segment_id',
            field=models.CharField(max_length=255, null=True),
        )
    ]



import json

from django.contrib import admin
from django import forms
from django.http import HttpResponseBadRequest, HttpResponse, JsonResponse
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.templatetags.static import static
from django.urls import reverse, path
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views.generic import CreateView

from nuntius.admin.fields import GenericModelChoiceField
from nuntius.celery import nuntius_celery_app
from nuntius.models import segment_cts, Campaign, MosaicoImage
from nuntius.tasks import send_campaign


class CampaignAdminForm(forms.ModelForm):
    segment = GenericModelChoiceField(querysets=lambda: [ct.model_class().objects.all() for ct in set(segment_cts())], required=False)

    def clean(self):
        cleaned_data = super().clean()
        segment = cleaned_data.get('segment')

        if segment:
            self.instance.segment = segment

        return cleaned_data

    class Meta:
        exclude = ('segment_content_type', 'segment_id')


class MosaicoImageUploadView(CreateView):
    model = MosaicoImage
    fields = ('file', )

    def image_dict(self, image):
        return {
            'name': image.file.name,
            'size': image.file.size,
            'url': self.request.build_absolute_uri(image.file.url),
            'deleteUrl': self.request.build_absolute_uri(image.file.url),
            'deleteType': 'DELETE',
            'thumbnailUrl': self.request.build_absolute_uri(image.file.thumbnail.url),
        }

    def form_invalid(self, form):
        return HttpResponseBadRequest()

    def form_valid(self, form):
        self.object = form.save()

        return JsonResponse({'files': [self.image_dict(self.object)]})

    def get(self, *args, **kwargs):
        return JsonResponse({'files': [self.image_dict(image) for image in MosaicoImage.objects.all()]})


@admin.register(Campaign)
class CampaignAdmin(admin.ModelAdmin):
    fieldsets = (
        (None, {
            'fields': ('name', 'message_from_name', 'message_from_email', 'message_reply_to_name',
                       'message_reply_to_email', 'message_subject')
        }),
        (_("Content"), {
            'fields': ('mosaico_buttons', 'message_content_text')
        }),
        (_("Sending"), {
            'fields': ('segment', 'segment_subscribers', 'status', 'send_button', 'sent_to', 'sent_ok',
                       'sent_bounced', 'sent_complained', 'sent_blocked', 'task_uuid', 'task_state')
        })
    )
    list_display = ('name', 'message_subject', 'segment', 'segment_subscribers', 'status', 'send_button', 'sent_to')
    list_filter = ('status',)
    form = CampaignAdminForm
    readonly_fields = ('segment_subscribers', 'status', 'send_button', 'mosaico_buttons', 'sent_to', 'sent_ok', 'sent_bounced',
                       'sent_complained', 'sent_blocked', 'task_uuid', 'task_state')

    def get_object(self, request, object_id, from_field=None):
        object = super().get_object(request, object_id, from_field=from_field)
        object.get_task_and_update_status()

        return object

    def segment_subscribers(self, instance):
        return instance.segment.get_subscribers_count()
    segment_subscribers.short_description = _("Subscribers")

    def sent_to(self, instance):
        return instance.get_sent_count()

    def sent_ok(self, instance):
        return instance.get_ok_count()
    sent_ok.short_description = _("Ok")

    def sent_bounced(self, instance):
        return instance.get_bounced_count()
    sent_bounced.short_description = _("Bounced")

    def sent_complained(self, instance):
        return instance.get_complained_count()
    sent_complained.short_description = _("Complained")

    def sent_blocked(self, instance):
        return instance.get_blocked_count()
    sent_blocked.short_description = _("Blocked")

    def task_state(self, instance):
        task = instance.get_task_and_update_status()
        if task is not None:
            return task[0]
        if instance.task_uuid is None:
            return '-'

        return _("Connection to celery failed")

    task_state.short_description = _("Send task state")

    def send_button(self, instance):
        if instance.pk is None:
            return mark_safe('-')
        if instance.status == Campaign.STATUS_SENDING:
            return format_html(
                '<a href="{}" class="button">' + _("Pause") + '</a>',
                reverse('admin:nuntius_campaign_pause', args=[instance.pk])
            )

        return format_html(
            '<a href="{}" class="button">' + _("Send") + '</a>',
            reverse('admin:nuntius_campaign_send', args=[instance.pk])
        )
    send_button.short_description = _("Send")

    def mosaico_buttons(self, instance):
        if instance.pk is None:
            return mark_safe('-')

        return format_html(
            '<a href="{}" class="button">' + _("Access the editor") + '</a> '\
            '<a href="{}" class="button">' + _("Preview result") + '</a>',
            reverse('admin:nuntius_campaign_mosaico', args=[instance.pk]) + '#' + static('nuntius/templates/versafix-1/template-versafix-1.html'),
            reverse('admin:nuntius_campaign_mosaico_preview', args=[instance.pk])
        )
    mosaico_buttons.short_description = _("HTML content")

    def get_urls(self):
        return [
            path('<pk>/send/', self.admin_site.admin_view(self.send_view), name='nuntius_campaign_send'),
            path('<pk>/pause/', self.admin_site.admin_view(self.pause_view), name='nuntius_campaign_pause'),
            path('<pk>/mosaico/', ensure_csrf_cookie(self.admin_site.admin_view(self.mosaico_view)), name='nuntius_campaign_mosaico'),
            path('<pk>/mosaico/preview/', self.admin_site.admin_view(self.mosaico_preview), name='nuntius_campaign_mosaico_preview'),
            path('<pk>/mosaico/save/', self.admin_site.admin_view(self.mosaico_save_view), name='nuntius_campaign_mosaico_save'),
            path('<pk>/mosaico/data/', self.admin_site.admin_view(self.mosaico_load_view), name='nuntius_campaign_mosaico_load'),
            path('<pk>/mosaico/upload/', self.admin_site.admin_view(MosaicoImageUploadView.as_view()), name='nuntius_campaign_mosaico_image_upload'),
        ] + super().get_urls()

    def send_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)
        if campaign.get_task_and_update_status() is not None:
            return redirect(reverse('admin:nuntius_campaign_change', args=[pk]))

        r = send_campaign.delay(pk)
        campaign.task_uuid = r.id
        campaign.save()

        return redirect(reverse('admin:nuntius_campaign_change', args=[pk]))

    def pause_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)
        if campaign.get_task_and_update_status() is None:
            return redirect(reverse('admin:nuntius_campaign_change', args=[pk]))

        nuntius_celery_app.control.revoke(str(campaign.task_uuid), terminate=True)

        return redirect(reverse('admin:nuntius_campaign_change', args=[pk]))

    def mosaico_view(self, request, pk):
        return TemplateResponse(request=request, template="nuntius/editor.html")

    def mosaico_preview(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)

        return HttpResponse(campaign.message_content_html)

    def mosaico_load_view(self, request, pk):
        campaign = Campaign.objects.get(pk=pk)

        return JsonResponse(json.loads(campaign.message_mosaico_data))

    def mosaico_save_view(self, request, pk):
        if request.method != 'POST':
            return HttpResponseBadRequest()

        html = request.POST.get('html')
        metadata = request.POST.get('metadata')
        content = request.POST.get('content')

        campaign = Campaign.objects.get(pk=pk)
        campaign.message_content_html = html
        campaign.message_mosaico_data = json.dumps({
            'metadata': json.loads(metadata),
            'content': json.loads(content)
        })
        campaign.save()

        return HttpResponse()



from django.test import TestCase

from nuntius.models import Campaign
from tests.models import TestSegment


class CampaignTestCase(TestCase):
    def test_can_attach_segment(self):
        segment = TestSegment.objects.create()
        campaign = Campaign.objects.create(name="Test campaign", segment=segment)

        campaign.refresh_from_db()
        self.assertEqual(campaign.segment, segment)



from functools import reduce

from celery.app.control import Inspect
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import fields, Q
from django.utils.translation import gettext as _
from stdimage import StdImageField

from nuntius.celery import nuntius_celery_app


def segment_cts_q():
    segment_cts_qs = [Q(app_label=models.split('.')[0], model=models.split('.')[1]) for models in
                      settings.NUNTIUS_SEGMENT_MODELS]
    return reduce(lambda q1, q2: q1 | q2, segment_cts_qs)


def segment_cts():
    return ContentType.objects.filter(segment_cts_q())


class EditableGenericForeignKey(GenericForeignKey):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.editable = True

    def save_object_data(self, value):
        return value

    def value_from_object(self, obj):
        return getattr(obj, self.name)


class Campaign(models.Model):
    STATUS_WAITING = 0
    STATUS_SENDING = 1
    STATUS_SENT = 2
    STATUS_ERROR = 3
    STATUS_CHOICES = (
        (STATUS_WAITING, _("Waiting")),
        (STATUS_SENDING, _("Sending")),
        (STATUS_SENT, _("Sent")),
        (STATUS_ERROR, _("Error"))
    )

    _task = None
    task_uuid = fields.UUIDField(_("Celery tasks identifier"), db_index=True, null=True, blank=True, default=None)

    name = fields.CharField(_("Name (invisible to subscribers)"), max_length=255)
    created = fields.DateTimeField(auto_now_add=True)
    updated = fields.DateTimeField(auto_now=True)

    message_from_name = fields.CharField(_("\"From\" name"), max_length=255, blank=True)
    message_from_email = fields.EmailField(_("\"From\" email address"), max_length=255)
    message_reply_to_name = fields.CharField(_("\"Reply to\" name"), max_length=255, blank=True)
    message_reply_to_email = fields.CharField(_("\"Reply to\" email address"), max_length=255, blank=True)
    message_subject = fields.CharField(_("Message subject line"), max_length=255, blank=True)
    message_mosaico_data = fields.TextField(_("Mosaico data"), blank=True)
    message_content_html = fields.TextField(_("Message content (HTML)"), blank=True)
    message_content_text = fields.TextField(_("Message content (text)"))

    segment_content_type = models.ForeignKey(ContentType, on_delete=models.PROTECT, limit_choices_to=segment_cts_q, null=True)
    segment_id = models.CharField(max_length=255, null=True)
    segment = EditableGenericForeignKey('segment_content_type', 'segment_id')

    status = fields.IntegerField(choices=STATUS_CHOICES, default=STATUS_WAITING)

    def get_task_and_update_status(self):
        # caching
        if self._task is not None:
            return self._task

        # no task known for this campaign
        if self.task_uuid is None:
            if self.status == Campaign.STATUS_SENDING:
                self.status = Campaign.STATUS_WAITING
                self.save(update_fields=['status'])
            return

        res = Inspect(app=nuntius_celery_app).query_task(self.task_uuid)

        # celery is down
        if res is None:
            return

        for host_tasks in res.values():
            if host_tasks.get(str(self.task_uuid)) is None:
                continue
            self.status = Campaign.STATUS_SENDING
            self.save(update_fields=['status'])
            self._task = host_tasks[str(self.task_uuid)]

            return host_tasks[str(self.task_uuid)]

        # celery is up but task is unkown
        self.task_uuid = None
        self.save(update_fields=['task_uuid'])

    def get_sent_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).exclude(result=CampaignSentEvent.RESULT_PENDING).count()

    def get_ok_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).filter(result=CampaignSentEvent.RESULT_OK).count()

    def get_bounced_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).filter(result=CampaignSentEvent.RESULT_BOUNCED).count()

    def get_complained_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).filter(result=CampaignSentEvent.RESULT_COMPLAINED).count()

    def get_blocked_count(self):
        return CampaignSentEvent.objects.filter(campaign=self).filter(result=CampaignSentEvent.RESULT_BLOCKED).count()


class BaseSegment:
    def get_display_name(self):
        raise NotImplementedError()

    def get_subscribers_queryset(self):
        raise NotImplementedError()

    def get_subscribers_count(self):
        raise NotImplementedError


class BaseSubscriber:
    STATUS_SUBSCRIBED = 1
    STATUS_UNSUBSCRIBED = 2
    STATUS_BOUNCED = 3
    STATUS_COMPLAINED = 4
    STATUS_CHOICES = (
        (STATUS_SUBSCRIBED, _("Subscribed")),
        (STATUS_UNSUBSCRIBED, _("Unsubscribed")),
        (STATUS_BOUNCED, _("Bounced")),
        (STATUS_COMPLAINED, _("Complained"))
    )

    def get_subscriber_status(self):
        if hasattr(self, 'subscriber_status'):
            return self.subscriber_status
        raise NotImplementedError()

    def get_subscriber_email(self):
        if hasattr(self, 'email'):
            return self.email

        raise NotImplementedError()

    def get_subscriber_data(self):
        return {
            'email': self.email
        }


class CampaignSentEvent(models.Model):
    RESULT_PENDING = 'P'
    RESULT_REFUSED = 'RE'
    RESULT_OK = 'OK'
    RESULT_BOUNCED = 'BC'
    RESULT_COMPLAINED = 'C'
    RESULT_BLOCKED = 'BL'
    RESULT_CHOICES = (
        RESULT_PENDING, _("Sending"),
        RESULT_PENDING, _("Refused by server"),
        RESULT_OK, _("Sent"),
        RESULT_BOUNCED, _("Bounced"),
        RESULT_COMPLAINED, _("Complained"),
        RESULT_BOUNCED, _("Bounced")
    )

    subscriber = models.ForeignKey(
        settings.NUNTIUS_SUBSCRIBER_MODEL,
        models.SET_NULL,
        verbose_name=_("Subscriber"),
        null=True,
        blank=True,
    )
    email = models.EmailField(_("Email adress at event time"))
    campaign = models.ForeignKey('Campaign', models.CASCADE, _("Campaign"))
    datetime = models.DateTimeField(auto_now_add=True)
    result = models.CharField(_("Operation result"), max_length=2, default=RESULT_PENDING)

    class Meta:
        unique_together = ('campaign', 'subscriber')


class MosaicoImage(models.Model):
    file = StdImageField(upload_to='mosaico', variations={'thumbnail': (90, 90)})
    created = fields.DateTimeField(auto_now_add=True)



from __future__ import absolute_import, unicode_literals
from celery import Celery

nuntius_celery_app = Celery('nuntius')

nuntius_celery_app.conf.task_default_routing_key = 'nuntius'
nuntius_celery_app.task_send_sent_event = True
nuntius_celery_app.conf.task_started = True

nuntius_celery_app.config_from_object(
    'django.conf:settings',
    namespace='NUNTIUS_CELERY'
)



# Generated by Django 2.0.6 on 2018-06-19 09:29

from django.db import migrations, models
import nuntius.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestSegment',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False, unique=True)),
            ],
            bases=(nuntius.models.BaseSegment, models.Model),
        ),
        migrations.CreateModel(
            name='TestSubscriber',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=255)),
                ('subscriber_status', models.IntegerField(choices=[(1, 'Subscribed'), (2, 'Unsubscribed'), (3, 'Bounced'), (4, 'Complained')])),
                ('segments', models.ManyToManyField(to='tests.TestSegment')),
            ],
            bases=(nuntius.models.BaseSubscriber, models.Model),
        ),
    ]



from django.contrib import admin

from tests.models import TestSubscriber, TestSegment


@admin.register(TestSubscriber)
class TestSubscriberAdmin(admin.ModelAdmin):
    list_display = ('email', 'subscriber_status')
    pass


@admin.register(TestSegment)
class TestSegmentAdmin(admin.ModelAdmin):
    pass



import re
from smtplib import SMTPServerDisconnected, SMTPRecipientsRefused, SMTPSenderRefused
from time import sleep

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.db import transaction
from django.template import Template, Context

from nuntius.celery import nuntius_celery_app
from nuntius.models import Campaign, BaseSubscriber, CampaignSentEvent


def replace_vars(string, data):
    var_regex = re.compile(r'\[([-a-zA-Z_]+)\]')
    context = Context(data)

    return Template(var_regex.sub(r'{{ \1 }}', string)).render(context=context)


@nuntius_celery_app.task()
def send_campaign(campaign_pk):
    campaign = Campaign.objects.get(pk=campaign_pk)
    campaign.status = Campaign.STATUS_SENDING
    campaign.save()

    if campaign.segment is None:
        model = settings.NUNTIUS_SUBSCRIBER_MODEL
        model_class = ContentType.objects.get(app_label=model.split('.')[0], model=model.split('.')[1].lower()).model_class()
        queryset = model_class.objects.all()
    else:
        queryset = campaign.segment.get_subscribers_queryset().exclude(campaignsentevent__campaign=campaign)

    try:
        with mail.get_connection() as connection:
            for subscriber in queryset.iterator():
                def send_subscriber(subscriber, retries=10):
                    if subscriber.get_subscriber_status() != BaseSubscriber.STATUS_SUBSCRIBED:
                        return

                    email = subscriber.get_subscriber_email()

                    (event, created) = CampaignSentEvent.objects.get_or_create(
                        campaign=campaign,
                        subscriber=subscriber,
                        email=email,
                    )

                    if event.result != CampaignSentEvent.RESULT_PENDING:
                        return

                    with transaction.atomic():
                        event = CampaignSentEvent.objects.select_for_update()\
                            .get(subscriber=subscriber, campaign=campaign)

                        if event.result != CampaignSentEvent.RESULT_PENDING:
                            return

                        subscriber_data = subscriber.get_subscriber_data()

                        message = EmailMultiAlternatives(
                            subject=campaign.message_subject,
                            body=replace_vars(campaign.message_content_text, subscriber_data),
                            from_email=campaign.message_from_email,
                            to=[email],
                            reply_to=campaign.message_reply_to_email,
                            connection=connection,
                        )
                        message.attach_alternative(replace_vars(campaign.message_content_html, subscriber_data), 'text/html')
                        try:
                            message.send()
                            event.result = CampaignSentEvent.RESULT_OK
                            event.save()
                        except SMTPServerDisconnected as e:
                            if retries == 0:
                                campaign.status = Campaign.STATUS_ERROR
                                campaign.save()
                                raise e
                            connection.close()

                            sleep(1/retries)

                            connection.open()
                            send_subscriber(subscriber, retries=retries-1)
                        except SMTPRecipientsRefused:
                            event.result = CampaignSentEvent.RESULT_BLOCKED
                            event.save()

                send_subscriber(subscriber)

            campaign.status = Campaign.STATUS_SENT
            campaign.save()
    except ConnectionError:
        campaign.status = Campaign.STATUS_ERROR
        campaign.save()




from PIL import Image, ImageDraw


def generate_placeholder(width, height):
    image = Image.new('RGB', (int(width), int(height)), '#707070')
    draw = ImageDraw.Draw(image)

    x = 0
    y = 0
    line_size = 40
    while y < height:
        draw.polygon([(x, y), (x + line_size, y), (x + line_size * 2, y + line_size), (x + line_size * 2, y + line_size * 2)], fill='#808080')
        draw.polygon([(x, y + line_size), (x + line_size, y + line_size * 2), (x, y + line_size * 2)], fill='#808080')
        x = x + line_size * 2
        if (x > width):
            x = 0
            y = y + line_size * 2

    (textwidth, textheight) = draw.textsize(f"{width} x {height}")
    draw.text(((int(width)-textwidth)/2, (int(height)-textheight)/2), f"{width} x {height}", (255, 255, 255))

    return image



from itertools import islice

from django.core.management import BaseCommand
from tqdm import tqdm

from nuntius.models import BaseSubscriber
from tests.models import TestSubscriber, TestSegment


class Command(BaseCommand):
    help = 'Loads plenty of fake emails for benchmark'

    def handle(self, *args, **options):
        (segment, created) = TestSegment.objects.get_or_create(id='fake_emails')
        fake_emails = TestSubscriber.objects.filter(email__startswith='fake_email')
        fake_emails.all().delete()

        batch_size = 1000

        objs = (TestSubscriber(
                email='fake_email' + str(i) + '@example.com',
                subscriber_status=BaseSubscriber.STATUS_SUBSCRIBED,
        ) for i in range(100000))

        with tqdm(total=100000) as pbar:
            while True:
                batch = list(islice(objs, batch_size))
                if not batch:
                    break
                TestSubscriber.objects.bulk_create(batch, batch_size)
                pbar.update(batch_size)

        Relation = TestSubscriber.segments.through
        Relation.objects.filter(testsegment=segment).delete()

        objs = (Relation(
            testsegment=segment,
            testsubscriber=test_subscriber
        ) for test_subscriber in fake_emails.all())

        with tqdm(total=100000) as pbar:
            while True:
                batch = list(islice(objs, batch_size))
                if not batch:
                    break
                Relation.objects.bulk_create(batch, batch_size)
                pbar.update(batch_size)





from django.db import models
from django.db.models import fields

from nuntius.models import BaseSubscriber, BaseSegment


class TestSegment(BaseSegment, models.Model):
    id = models.CharField(max_length=255, unique=True, primary_key=True)

    def get_display_name(self):
        return self.id

    def get_subscribers_queryset(self):
        return TestSubscriber.objects.filter(segments__id=self.id)

    def get_subscribers_count(self):
        return self.get_subscribers_queryset().filter(subscriber_status=BaseSubscriber.STATUS_SUBSCRIBED).count()

    def __str__(self):
        return self.get_display_name()


class TestSubscriber(BaseSubscriber, models.Model):
    email = fields.EmailField(max_length=255)
    subscriber_status = fields.IntegerField(choices=BaseSubscriber.STATUS_CHOICES)

    segments = models.ManyToManyField('TestSegment')
